/*
    This file is part of GNU APL, a free implementation of the
    ISO/IEC Standard 13751, "Programming Language APL, Extended"

    Copyright (C) 2008-2014  Dr. Jürgen Sauermann

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// read-only system variables
ro_sv_def(Quad_AI )
ro_sv_def(Quad_ARG)
ro_sv_def(Quad_AV )
ro_sv_def(Quad_EM )
ro_sv_def(Quad_ET )
ro_sv_def(Quad_LC )
ro_sv_def(Quad_PT )   /* derived from ⎕PP */
ro_sv_def(Quad_SVE)
ro_sv_def(Quad_TC )
ro_sv_def(Quad_TS )
ro_sv_def(Quad_UL )
ro_sv_def(Quad_WA )
#undef ro_sv_def

// read-write system variables
rw_sv_def(Quad_CT )
rw_sv_def(Quad_FC )
rw_sv_def(Quad_IO )
rw_sv_def(Quad_L  )
rw_sv_def(Quad_LX )
rw_sv_def(Quad_PP )
rw_sv_def(Quad_PR )
rw_sv_def(Quad_PS )
rw_sv_def(Quad_PW )
rw_sv_def(Quad_R  )
rw_sv_def(Quad_RL )
rw_sv_def(Quad_SYL)
rw_sv_def(Quad_TZ )
rw_sv_def(Quad_X  )

rw_sv_def(ALPHA   )
rw_sv_def(ALPHA_U )
rw_sv_def(CHI     )
rw_sv_def(LAMBDA  )
rw_sv_def(OMEGA   )
rw_sv_def(OMEGA_U )
#undef rw_sv_def

// system functions
#ifdef sf_def
  sf_def(Quad_AF)
  sf_def(Quad_AT)
  sf_def(Quad_CR)
  sf_def(Quad_DL)
  sf_def(Quad_EA)
  sf_def(Quad_EC)
  sf_def(Quad_ENV)
  sf_def(Quad_ES)
  sf_def(Quad_FX)
  sf_def(Quad_INP)
  sf_def(Quad_NA)
  sf_def(Quad_NC)
  sf_def(Quad_NL)
  sf_def(Quad_SI)
  sf_def(Quad_SVC)
  sf_def(Quad_SVO)
  sf_def(Quad_SVQ)
  sf_def(Quad_SVR)
  sf_def(Quad_SVS)
  sf_def(Quad_TF)
  sf_def(Quad_UCS)
# undef sf_def
#endif



